using System.Net;
using DotNext;
using WebProtocol;

namespace JtagClient;

/// <summary>
/// Global Constant Jtag Address
/// </summary>
public static class JtagAddr
{
    /// <summary>
    /// Jtag State Reg
    /// </summary>
    public const UInt32 STATE = 0x10_00_00_00;
    /// <summary>
    /// Jtag Read Reg
    /// </summary>
    public const UInt32 READ_DATA = 0x10_00_00_01;
    /// <summary>
    /// Jtag Write Reg
    /// </summary>
    public const UInt32 WRITE_DATA = 0x10_00_00_02;
    /// <summary>
    /// Jtag Write Command
    /// </summary>
    public const UInt32 WRITE_CMD = 0x10_00_00_03;
}

/// <summary>
/// Jtag 状态寄存器的掩码及其确认函数
/// </summary>
public static class JtagState
{

    /*
    [0]    移位数据读fifo清空，1为有效，清空后自动置0
    [1]    移位数据fifo读入口-空标识，只读，对JTAG_STATE_REG的写不改变其值
    [2]    移位数据fifo读入口-满标识，只读，对JTAG_STATE_REG的写不改变其值
    [7:3]  保留

    [8]    移位数据写fifo清空，1为有效，清空后自动置0
    [9]    移位数据fifo写入口-空标识，只读，对JTAG_STATE_REG的写不改变其值
    [10]   移位数据fifo写入口-满标识，只读，对JTAG_STATE_REG的写不改变其值
    [15:11]保留

    [16]   移位命令写fifo清空，1为有效，清空后自动置0
    [17]   移位命令fifo写入口-空标识，只读，对JTAG_STATE_REG的写不改变其值
    [18]   移位命令fifo写入口-满标识，只读，对JTAG_STATE_REG的写不改变其值
    [23:19]保留

    [24]   CMD执行完毕标识，只读，对JTAG_STATE_REG的写不改变其值
    */

    /// <summary>
    /// 移位数据读fifo清空，1为有效，清空后自动置0，实际一直为零
    /// </summary>
    public const UInt32 READ_DATA_FIFO_CLEAR = 0b0001 << 0;
    /// <summary>
    /// 移位数据fifo读入口-空标识，只读，对JTAG_STATE_REG的写不改变其值
    /// </summary>
    public const UInt32 READ_DATA_FIFO_EMPTY = 0b0010 << 0;
    /// <summary>
    /// 移位数据fifo读入口-满标识，只读，对JTAG_STATE_REG的写不改变其值
    /// </summary>
    public const UInt32 READ_DATA_FIFO_AFULL = 0b0100 << 0;
    /// <summary>
    /// 移位数据读fifo
    /// </summary>
    public const UInt32 READ_DATA_FIFO = 0b0111 << 0;

    /// <summary>
    /// 移位数据写fifo清空，1为有效，清空后自动置0
    /// </summary>
    public const UInt32 WRITE_DATA_FIFO_CLEAR = 0b0001 << 8;
    /// <summary>
    /// 移位数据fifo写入口-空标识，只读，对JTAG_STATE_REG的写不改变其值
    /// </summary>
    public const UInt32 WRITE_DATA_FIFO_EMPTY = 0b0010 << 8;
    /// <summary>
    /// 移位数据fifo写入口-满标识，只读，对JTAG_STATE_REG的写不改变其值
    /// </summary>
    public const UInt32 WRITE_DATA_FIFO_AFULL = 0b0100 << 8;
    /// <summary>
    /// 移位数据写fifo
    /// </summary>
    public const UInt32 WRITE_DATA_FIFO = 0b0111 << 8;

    /// <summary>
    /// 移位命令写fifo清空，1为有效，清空后自动置0
    /// </summary>
    public const UInt32 WRITE_CMD_FIFO_CLEAR = 0b0001 << 16;
    /// <summary>
    /// 移位命令fifo写入口-空标识，只读，对JTAG_STATE_REG的写不改变其值
    /// </summary>
    public const UInt32 WRITE_CMD_FIFO_EMPTY = 0b0010 << 16;
    /// <summary>
    /// 移位命令fifo写入口-满标识，只读，对JTAG_STATE_REG的写不改变其值
    /// </summary>
    public const UInt32 WRITE_CMD_FIFO_AFULL = 0b0100 << 16;
    /// <summary>
    /// 移位命令写fifo
    /// </summary>
    public const UInt32 WRITE_CMD_FIFO = 0b0111 << 16;

    /// <summary>
    /// CMD执行完毕标识，只读，对JTAG_STATE_REG的写不改变其值
    /// </summary>
    public const UInt32 CMD_EXEC_FINISH = 0b0001 << 24;
    /// <summary>
    /// 全部FIFO
    /// </summary>
    public const UInt32 ALL_FIFO = READ_DATA_FIFO | WRITE_DATA_FIFO | WRITE_CMD_FIFO;
    /// <summary>
    /// 全部寄存器
    /// </summary>
    public const UInt32 ALL_REG = READ_DATA_FIFO | WRITE_DATA_FIFO | WRITE_CMD_FIFO | CMD_EXEC_FINISH;

}

/// <summary>
/// The Command bits of Jtag
/// </summary>
public static class JtagCmd
{

    /// <summary>
    /// The length of JTAG_DR_XXXX
    /// </summary>
    public const UInt32 LEN_JTAG_DR = 10;
    /// <summary>
    /// 旁路指令
    /// </summary>
    public const UInt32 JTAG_DR_BYPASS = 0b1111111111;
    /// <summary>
    /// 采样指令
    /// </summary>
    public const UInt32 JTAG_DR_SAMPLE = 0b1010000000;
    /// <summary>
    /// 预装指令
    /// </summary>
    public const UInt32 JTAG_DR_PRELOAD = 0b1010000000;
    /// <summary>
    /// 外测试指令
    /// </summary>
    public const UInt32 JTAG_DR_EXTEST = 0b1010000001;
    /// <summary>
    /// 内测试指令
    /// </summary>
    public const UInt32 JTAG_DR_INTEST = 0b1010000010;
    /// <summary>
    /// 标识指令
    /// </summary>
    public const UInt32 JTAG_DR_IDCODE = 0b1010000011;
    /// <summary>
    /// 高阻指令
    /// </summary>
    public const UInt32 JTAG_DR_HIGHZ = 0b1010000101;
    /// <summary>
    /// 复位指令
    /// </summary>
    public const UInt32 JTAG_DR_JRST = 0b1010001010;
    /// <summary>
    /// 配置指令
    /// </summary>
    public const UInt32 JTAG_DR_CFGI = 0b1010001011;
    /// <summary>
    /// 回读指令
    /// </summary>
    public const UInt32 JTAG_DR_CFGO = 0b1010001100;
    /// <summary>
    /// 唤醒指令
    /// </summary>
    public const UInt32 JTAG_DR_JWAKEUP = 0b1010001101;
    /// <summary>
    /// 读UID指令
    /// </summary>
    public const UInt32 JTAG_DR_READ_UID = 0b0101001100;
    /// <summary>
    /// 读状态寄存器指令
    /// </summary>
    public const UInt32 JTAG_DR_RDSR = 0b0101011001;



    /// <summary>
    /// The length of CMD_JTAG_XXXX_XXXX
    /// </summary>
    public const UInt32 LEN_CMD_JTAG = 4;
    /// <summary>
    /// 设定JTAG默认状态为TEST_LOGIC_RESET态 （JTAG复位）
    /// </summary>
    public const UInt32 CMD_JTAG_CLOSE_TEST = 0b0000;
    /// <summary>
    /// 设定JTAG默认状态为RUN_TEST_IDLE态    （JTAG空闲）
    /// </summary>
    public const UInt32 CMD_JTAG_RUN_TEST = 0b0001;
    /// <summary>
    /// JTAG进入SHIFTIR循环cycle_num次后回到设定的默认状态
    /// </summary>
    public const UInt32 CMD_JTAG_LOAD_IR = 0b0010;
    /// <summary>
    /// JTAG进入SHIFTDR循环cycle_num次后回到设定的默认状态，同时输入shift_in
    /// </summary>
    public const UInt32 CMD_JTAG_LOAD_DR_CAREI = 0b0011;
    /// <summary>
    /// JTAG进入SHIFTDR循环cycle_num次后回到设定的默认状态，同时输出shift_out
    /// </summary>
    public const UInt32 CMD_JTAG_LOAD_DR_CAREO = 0b0100;
    /// <summary>
    /// JTAG进入RUN_TEST_IDLE态循环cycle_num次
    /// </summary>
    public const UInt32 CMD_JTAG_IDLE_DELAY = 0b0101;
}

/// <summary>
/// Jtag控制器
/// </summary>
public class Jtag
{
    private static readonly NLog.Logger logger = NLog.LogManager.GetCurrentClassLogger();

    readonly int timeout;

    readonly int port;
    readonly string address;
    private IPEndPoint ep;

    /// <summary>
    /// Jtag构造函数
    /// </summary>
    /// <param name="address">目标IP地址</param>
    /// <param name="port">目标UDP端口</param>
    /// <param name="timeout">超时时间</param>
    public Jtag(string address, int port, int timeout = 2000)
    {
        this.address = address;
        this.port = port;
        this.ep = new IPEndPoint(IPAddress.Parse(address), port);
        this.timeout = timeout;
    }

    async ValueTask<Result<uint>> ReadFIFO(uint devAddr)
    {
        var ret = false;
        var opts = new SendAddrPackOptions();

        opts.BurstType = BurstType.FixedBurst;
        opts.BurstLength = 0;
        opts.CommandID = 0;
        opts.Address = devAddr;

        // Read Jtag State Register
        opts.IsWrite = false;
        ret = await UDPClientPool.SendAddrPackAsync(ep, new SendAddrPackage(opts));
        if (!ret) return new(new Exception("Send Address Package Failed!"));

        // Wait for Read Ack
        if (!MsgBus.IsRunning)
            return new(new Exception("Message Bus not Working!"));

        var retPack = await MsgBus.UDPServer.WaitForDataAsync(address, port);
        if (!retPack.IsSuccessful || !retPack.Value.IsSuccessful)
            return new(new Exception("Send address package failed"));

        var retPackOpts = retPack.Value.Options;
        if (retPackOpts.Data is null)
            return new(new Exception($"Data is Null, package: {retPackOpts.ToString()}"));

        var retPackLen = retPackOpts.Data.Length;
        if (retPackLen != 4)
            return new(new Exception($"RecvDataPackage BodyData Length not Equal to 4: Total {retPackLen} bytes"));

        return Convert.ToUInt32(Common.Number.BytesToNumber(retPackOpts.Data).Value);
    }

    public async ValueTask<Result<RecvDataPackage>> WriteFIFO(UInt32 devAddr, UInt32 data)
    {
        var ret = false;
        var opts = new SendAddrPackOptions();


        opts.BurstType = BurstType.FixedBurst;
        opts.BurstLength = 0;
        opts.CommandID = 0;
        opts.Address = devAddr;

        // Write Jtag State Register
        opts.IsWrite = true;
        ret = await UDPClientPool.SendAddrPackAsync(ep, new SendAddrPackage(opts));
        if (!ret) return new(new Exception("Send 1st address package failed!"));
        // Send Data Package
        ret = await UDPClientPool.SendDataPackAsync(ep,
                new SendDataPackage(Common.Number.NumberToBytes(data, 4).Value));
        if (!ret) return new(new Exception("Send data package failed!"));

        // Check Msg Bus
        if (!MsgBus.IsRunning)
            return new(new Exception("Message bus not working!"));
        // Wait for Write Ack
        var udpWriteAck = await MsgBus.UDPServer.WaitForAckAsync(address, port);
        if (!udpWriteAck.IsSuccessful) return new(udpWriteAck.Error);
        else if (!udpWriteAck.Value.IsSuccessful)
            return new(new Exception("Send address package failed"));

        // Read Jtag State Register
        opts.IsWrite = false;
        opts.Address = JtagAddr.STATE;
        ret = await UDPClientPool.SendAddrPackAsync(ep, new SendAddrPackage(opts));
        if (!ret) return new(new Exception("Send 2rd Address Package Failed!"));
        // Wait for Read Data
        var udpDataResp = await MsgBus.UDPServer.WaitForDataAsync(address, port);
        if (!udpDataResp.IsSuccessful) return new(udpDataResp.Error);
        else if (!udpDataResp.Value.IsSuccessful)
            return new(new Exception("Send address package failed"));

        return udpDataResp.Value;
    }

    async ValueTask<Result<RecvDataPackage>> WriteFIFO(UInt32 devAddr, byte[] dataArray)
    {
        var ret = false;
        var opts = new SendAddrPackOptions();


        opts.BurstType = BurstType.FixedBurst;
        opts.CommandID = 0;
        opts.Address = devAddr;

        // Check Msg Bus
        if (!MsgBus.IsRunning)
            return new(new Exception("Message bus not working!"));

        var writeTimes = dataArray.Length / (256 * (32 / 8)) + 1;
        for (var i = 0; i < writeTimes; i++)
        {
            var isLastData = i == writeTimes - 1;
            // Write Jtag State Register
            opts.IsWrite = true;
            opts.BurstLength = isLastData ? (byte)(dataArray.Length - i * (256 * (32 / 8)) - 1) : (byte)255;
            ret = await UDPClientPool.SendAddrPackAsync(ep, new SendAddrPackage(opts));
            if (!ret) return new(new Exception("Send 1st address package failed!"));
            // Send Data Package
            ret = await UDPClientPool.SendDataPackAsync(ep,
                    new SendDataPackage(
                        isLastData ?
                        dataArray[(i * (256 * (32 / 8)))..] :
                        dataArray[(i * (256 * (32 / 8)))..((i + 1) * (256 * (32 / 8)) - 1)]
                    )
                );
            if (!ret) return new(new Exception("Send data package failed!"));

            // Wait for Write Ack
            var udpWriteAck = await MsgBus.UDPServer.WaitForAckAsync(address, port);
            if (!udpWriteAck.IsSuccessful) return new(udpWriteAck.Error);
            else if (!udpWriteAck.Value.IsSuccessful)
                return new(new Exception("Send address package failed"));
        }

        // Read Jtag State Register
        opts.IsWrite = false;
        opts.BurstLength = 0;
        opts.Address = JtagAddr.STATE;
        ret = await UDPClientPool.SendAddrPackAsync(ep, new SendAddrPackage(opts));
        if (!ret) return new(new Exception("Send 2rd Address Package Failed!"));
        // Wait for Read Data
        var udpDataResp = await MsgBus.UDPServer.WaitForDataAsync(address, port);
        if (!udpDataResp.IsSuccessful) return new(udpDataResp.Error);
        else if (!udpDataResp.Value.IsSuccessful)
            return new(new Exception("Send address package failed"));

        return udpDataResp.Value;
    }

    async ValueTask<Result<bool>> WriteFIFO
        (UInt32 devAddr, UInt32 data, UInt32 result, UInt32 resultMask = 0xFF_FF_FF_FF)
    {
        var ret = false;
        var retPack = await WriteFIFO(devAddr, data);
        if (!retPack.IsSuccessful) return new(retPack.Error);

        if (retPack.Value.Options.Data is null)
            return new(new Exception($"Data is Null, package: {retPack.Value.Options.ToString()}"));

        var retPackLen = retPack.Value.Options.Data.Length;
        if (retPackLen != 4)
            return new(new Exception($"RecvDataPackage BodyData Length not Equal to 4: Total {retPackLen} bytes"));

        if (Common.Number.BitsCheck(
                Common.Number.BytesToNumber(retPack.Value.Options.Data).Value, result, resultMask))
            ret = true;

        return ret;
    }

    async ValueTask<Result<bool>> WriteFIFO
        (UInt32 devAddr, byte[] data, UInt32 result, UInt32 resultMask = 0xFF_FF_FF_FF)
    {
        var ret = false;
        var retPack = await WriteFIFO(devAddr, data);

        if (retPack.Value.Options.Data is null)
            return new(new Exception($"Data is Null, package: {retPack.Value.Options.ToString()}"));

        var retPackLen = retPack.Value.Options.Data.Length;
        if (retPackLen != 4)
            return new(new Exception($"RecvDataPackage BodyData Length not Equal to 4: Total {retPackLen} bytes"));

        if (Common.Number.BitsCheck(
                Common.Number.BytesToNumber(retPack.Value.Options.Data).Value, result, resultMask))
            ret = true;

        return ret;
    }


    async ValueTask<Result<bool>> ClearAllRegisters()
    {
        return await WriteFIFO(JtagAddr.STATE, 0xFF_FF_FF_FF, 0x01_02_02_02, JtagState.ALL_REG);
    }

    async ValueTask<Result<bool>> ClearWriteDataReg()
    {
        return await WriteFIFO(JtagAddr.STATE, 0x00_00_11_00, 0x01_00_02_00, JtagState.WRITE_DATA_FIFO | JtagState.CMD_EXEC_FINISH);
    }

    async ValueTask<Result<bool>> CloseTest()
    {
        return await WriteFIFO(
                JtagAddr.WRITE_CMD,
                Common.Number.MultiBitsToNumber(JtagCmd.CMD_JTAG_CLOSE_TEST, JtagCmd.LEN_CMD_JTAG, 0, 28).Value,
                0x01_00_00_00, JtagState.CMD_EXEC_FINISH);
    }

    async ValueTask<Result<bool>> RunTest()
    {
        return await WriteFIFO(
                JtagAddr.WRITE_CMD,
                Common.Number.MultiBitsToNumber(JtagCmd.CMD_JTAG_RUN_TEST, JtagCmd.LEN_CMD_JTAG, 0, 28).Value,
                0x01_00_00_00, JtagState.CMD_EXEC_FINISH);
    }

    async ValueTask<Result<bool>> IdleDelay(UInt32 milliseconds)
    {
        if (milliseconds > Math.Pow(2, 28)) return new(new Exception("Timespan is over 2^28 milliseconds"));

        return await WriteFIFO(
                JtagAddr.WRITE_CMD,
                Common.Number.MultiBitsToNumber(JtagCmd.CMD_JTAG_IDLE_DELAY, JtagCmd.LEN_CMD_JTAG, milliseconds, 28).Value,
                0x01_00_00_00, JtagState.CMD_EXEC_FINISH);
    }

    async ValueTask<Result<bool>> ExecRDCmd(uint command)
    {
        {
            var ret = await WriteFIFO(
                JtagAddr.WRITE_DATA,
                Common.Number.MultiBitsToNumber(0, 22, command, JtagCmd.LEN_JTAG_DR).Value, 0, 0);

            if (!ret.IsSuccessful) return new(ret.Error);
            else if (!ret.Value) return new(new Exception("Write JTAG_DR_IDCODE Failed"));
        }
        {
            var ret = await WriteFIFO(
                JtagAddr.WRITE_CMD,
                Common.Number.MultiBitsToNumber(JtagCmd.CMD_JTAG_LOAD_IR, JtagCmd.LEN_CMD_JTAG, 10, 28).Value,
                0x01_00_00_00, JtagState.CMD_EXEC_FINISH);

            if (!ret.IsSuccessful) return new(ret.Error);
            else if (!ret.Value) return new(new Exception("Write CMD_JTAG_LOAD_IR Failed"));
        }
        return await ClearWriteDataReg();
    }

    async ValueTask<Result<bool>> LoadDRCareInput(byte[] bytesArray)
    {
        var bytesLen = ((uint)(bytesArray.Length * 8));
        if (bytesLen > Math.Pow(2, 28)) return new(new Exception("Length is over 2^(28 - 3)"));

        {
            var ret = await WriteFIFO(
                    JtagAddr.WRITE_CMD,
                    Common.Number.MultiBitsToNumber(JtagCmd.CMD_JTAG_LOAD_DR_CAREI, JtagCmd.LEN_CMD_JTAG, bytesLen, 28).Value,
                    0x01_00_00_00, JtagState.CMD_EXEC_FINISH);

            if (!ret.IsSuccessful) return new(ret.Error);
            else if (!ret.Value) return new(new Exception("Write CMD_JTAG_LOAD_DR_CAREI Failed"));
        }
        {
            var ret = await WriteFIFO(JtagAddr.WRITE_DATA, bytesArray, 0x01_00_00_00, JtagState.CMD_EXEC_FINISH);

            if (!ret.IsSuccessful) return new(ret.Error);
            else if (!ret.Value) return new(new Exception("Write Data Failed"));
        }
        return true;
    }

    async ValueTask<Result<UInt32>> LoadDRCareOutput(UInt32 bytesLen)
    {
        if (bytesLen > Math.Pow(2, 28)) return new(new Exception("Length is over 2^(28 - 3)"));

        var ret = await WriteFIFO(
                JtagAddr.WRITE_CMD,
                Common.Number.MultiBitsToNumber(JtagCmd.CMD_JTAG_LOAD_DR_CAREO, JtagCmd.LEN_CMD_JTAG, 8 * bytesLen, 28).Value,
                0x01_00_00_00, JtagState.CMD_EXEC_FINISH);

        if (ret.Value)
            return await ReadFIFO(JtagAddr.READ_DATA);
        else
            return new(new Exception("LoadDRCareo Failed!"));
    }

    public async ValueTask<Result<uint>> ReadIDCode()
    {
        // Clear Data
        await MsgBus.UDPServer.ClearUDPData(this.address);

        logger.Trace($"Clear up udp server {this.address} receive data");

        Result<bool> ret;

        ret = await ClearAllRegisters();
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag Clear All Registers Failed"));

        ret = await RunTest();
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag Run Test Failed"));

        ret = await ExecRDCmd(JtagCmd.JTAG_DR_IDCODE);
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag Execute Command JTAG_DR_IDCODE Failed"));

        ret = await ClearWriteDataReg();
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag Clear Write Registers Failed"));

        var retData = await LoadDRCareOutput(4);
        if (!retData.IsSuccessful)
        {
            return new(new Exception("Get ID Code Failed"));
        }

        return retData.Value;
    }

    public async ValueTask<Result<bool>> DownloadBitstream(byte[] bitstream)
    {
        // Clear Data
        await MsgBus.UDPServer.ClearUDPData(this.address);

        logger.Trace($"Clear up udp server {this.address} receive data");

        Result<bool> ret;

        ret = await CloseTest();
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag Close Test Failed"));

        ret = await RunTest();
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag Run Test Failed"));

        logger.Trace("Jtag initialize");

        ret = await ExecRDCmd(JtagCmd.JTAG_DR_JRST);
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag Execute Command JTAG_DR_JRST Failed"));

        ret = await RunTest();
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag Run Test Failed"));

        ret = await ExecRDCmd(JtagCmd.JTAG_DR_CFGI);
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag Execute Command JTAG_DR_CFGI Failed"));

        logger.Trace("Jtag ready to write bitstream");

        ret = await IdleDelay(75000);
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag IDLE Delay Failed"));

        ret = await LoadDRCareInput(bitstream);
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag Load Data Failed"));

        logger.Trace("Jtag write bitstream");

        ret = await CloseTest();
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag Close Test Failed"));

        ret = await RunTest();
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag Run Test Failed"));

        ret = await ExecRDCmd(JtagCmd.JTAG_DR_JWAKEUP);
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag Execute Command JTAG_DR_JWAKEUP Failed"));

        logger.Trace("Jtag reset device");

        ret = await IdleDelay(1000);
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag IDLE Delay Failed"));

        ret = await CloseTest();
        if (!ret.IsSuccessful) return new(ret.Error);
        else if (!ret.Value) return new(new Exception("Jtag Close Test Failed"));
        logger.Trace("Jtag download bitstream successfully");

        return true;
    }

}
